from __future__ import annotations

from dissect.volume.disk.schemes.apm import APM
from dissect.volume.disk.schemes.bsd import BSD
from dissect.volume.disk.schemes.gpt import GPT
from dissect.volume.disk.schemes.mbr import MBR

__all__ = [
    "APM",
    "BSD",
    "GPT",
    "MBR",
]
